class yambopyenv():
    YAMBO = "yambo"
    P2Y = "p2y"
    E2Y = "e2y"
    YPP = "ypp"
    SCHEDULER = "bash"
    YAMBO_RT = "yambo_rt"
    YPP_RT = "ypp_rt"
    YAMBO_NL = "yambo_nl"
    YPP_NL = "ypp_nl"