# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
"""
submodule to handle databases from yambo
Here we will add classes to read and manipulate the netcdf databases that yambo creates
This will allow the user to create his own scripts to manipulate the data
"""
