# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
"""
Scripts to manipulate Quantum Espresso input files

Also able to read output files in xml format (datafile.xml or datafile-schema.xml)

"""
import os

class qepyenv():
    PW = "pw.x"
    PH = "ph.x"
    DYNMAT = "dynmat.x"
    PSEUDODIR = os.path.join(os.path.dirname(__file__),'data','pseudos')
    CONV_THR = 1e-8

from qepy.lattice import *
from qepy.xml import *
from qepy.bravais import *
from qepy.pw import *
from qepy.pwxml import *
from qepy.projwfc import *
from qepy.projwfcxml import *
from qepy.ph import *
from qepy.dynmat import *
from qepy.matdyn import *
from qepy.unfolding import *
from qepy.supercell import *
from qepy.upf_interface.ppupf import *
from qepy.spin_texture import *
