rm -rf relax scf nscf database nscf_double phonon database_double ip bands gw gw_conv gw-xi gw-zeros elphon gw_cutoff bse_cutoff bse_cutoff_cut gw_par bse bse_conv bse_par bse_par_bug gw_bse relax.log scf.log nscf.log nscf_double.log yambo_bse.log yambo_gw.log yambo.log p2y.log yambo_par_bse.log jobs.sh *.json bse_run em1s.pdf em1s.dat analyse_gw_conv analyse_bse_conv absorptionspectra.json *.png *.pdf l-p2y r_setup run.sh gw_pp_par nscf_flow bse_flow r_setup* gw+bse run.sh bse_calc l_p2y l_p2y_* l_setup l_setup_*
