# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
"""
submodule to handle input and output files of yambo.
by output here we mean ``o-``, ``r-`` and ``l-`` files
the databases are handled in the db module
"""
