# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
#
"""
submodule with classes to read BSE optical absorption spectra claculations
and information about the excitonic states
"""
